(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/util/Dot.sml,v
   Comment: A very simple Dot (graphwiz) toolkit. Contains incomplete
	    representation of DOT language and a "pretty"-printer for it.
	    Graphviz is developed at AT&T labs (and distributed under their
	    license).
   Release: scope-0-15_final 

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Dot.sml,v 1.11 2003/07/07 19:02:19 wasowski Exp
*)

structure Dot = struct

	type strict = bool
	type id = string

	datatype style     = Dashed | Dotted | Filled | Solid 
	datatype shape     = Box | Circle | Diamond | Point | Invis
	datatype color     = Black | LightGray | White
	datatype port      = N | NE | E | SE | S | SW | W | NW
	datatype arrowhead = None | Normal

	datatype attr = ArrowHead     of arrowhead
		      | Color         of color
		      | Compound      of bool
		      | HeadPort      of port
		      | Height        of real
		      | FillColor     of color
		      | FixedSize     of bool
		      | FontColor     of color
		      | FontName      of string
		      | FontSize      of int
		      | Label         of id
		      | LabelFontName of string
		      | LabelFontSize of int
		      | LHead         of string
		      | LTail	      of string
		      | Style         of style 
		      | Shape         of shape
		      | TailPort      of port
		      | Width         of real

	type node = { 
		
		id :id,
		attrs :attr list
	}

	type edge = {
		
		head  :id,
		tail  :id,
		dir   :bool,
		attrs :attr list
	}

	datatype stmt = Node      of node
		      | Edge      of edge 
		      | NodeAttr  of attr list
		      | EdgeAttr  of attr list
		      | GraphAttr of attr list
		      | Subgraph  of graph

	withtype graph = { 
			cluster :bool,
			id :id,
			stmts : stmt list 
		}

	infix 0 -->
	infix 0 ---

	fun a --> b = fn attrs => Edge { head=b, tail=a, dir=true, attrs = attrs }
	fun a --- b = fn attrs => Edge { head=b, tail=a, dir=false, attrs = attrs }

	type topgraph = bool * bool * graph (* strict, digraph, rest *)

	type dot = topgraph list

	(* pretty printer *)

	fun prstyle Dashed = "dashed"
	  | prstyle Dotted = "dotted"
	  | prstyle Filled = "filled"
	  | prstyle Solid  = "solid"

	fun prshape Box     = "box"
	  | prshape Diamond = "diamond"
	  | prshape Point   = "point"
	  | prshape Circle  = "circle"
	  | prshape Invis   = "invis"

	fun prarrowhead Normal = "normal"
	  | prarrowhead None   = "none"

	fun prcolor Black     = "black"
	  | prcolor LightGray = "lightgray"
	  | prcolor White     = "white"

	fun prport N  = "n"
	  | prport NE = "ne"
	  | prport E  = "e"
	  | prport SE = "se"
	  | prport S  = "s"
	  | prport SW = "sw"
	  | prport W  = "w"
	  | prport NW = "nw"

	fun prbool true  = "true"
	  | prbool false = "false"
	
	fun decluster clusters name = 
		if List.exists (fn c=>c=name) clusters 
		then "cluster" ^ name else name

	(* constructs a field from name and value*)
	fun mkf(name :string) (value :string) cont :string list =
		" " :: value :: "=" :: name :: cont
	(* constructs a quoted field from name and value*)
	fun mkq(name :string) (value :string) cont :string list =
		"\" " :: value :: "=\"" :: name :: cont

	fun prattr clusters (a :attr) (cont :string list) :string list=
		 case a of ArrowHead a => mkf "arrowhead" (prarrowhead a)   cont
			 | Color    c  => mkf "color"     (prcolor c)       cont
			 | Compound b  => mkf "compound"  (prbool b)        cont
			 | HeadPort p  => mkf "headport"  (prport p)        cont
			 | Height h    => mkf "height"    (Real.toString h) cont
			 | FillColor c => mkf "fillcolor" (prcolor c)       cont
			 | FixedSize b => mkf "fixedsize" (prbool b)        cont
			 | FontColor c => mkf "fontcolor" (prcolor c)       cont
			 | FontName n  => mkq "fontname"  n                 cont
			 | FontSize n  => mkf "fontsize"  (Int.toString n)  cont
			 | Label l     => mkq "label"     l                 cont
			 | LabelFontName n => 
				mkq "labelfontname" n cont
			 | LabelFontSize n => 
				mkf "labelfontsize" (Int.toString n) cont
			 | LHead n     => 
				mkq "lhead" (decluster clusters n) cont
			 | LTail n     => 
				mkq "ltail" (decluster clusters n) cont
			 | Style s     => mkf "style"     (prstyle s)       cont
			 | Shape s     => mkf "shape"     (prshape s)       cont
			 | TailPort p  => mkf "tailport"  (prport p)        cont
			 | Width r     => mkf "width"     (Real.toString r) cont 

	fun prattrs clusters name []    cont = "\n" :: name ::cont
	  | prattrs clusters name attrs cont = 
		"]\n" :: foldl (fn (a,c) => (prattr clusters) a c) 
			(" [" :: name :: cont) attrs


	(* TODO: would be nice to have more efficient representation *)
	(* of clusters                                               *)
	fun predge clusters {tail=b, head=a, attrs, dir} cont :string list =
		let val tail = decluster clusters b
		    val head = decluster clusters a
		    val ope  = if dir then " -> " else " -- "
		    val arr  = String.concat ["\"",tail,"\"", ope, "\"", head, "\""]
		in prattrs clusters arr attrs cont end

	fun prstmt clusters (s,cont) :string list =
		case s of
			  Edge e          => predge clusters e     cont
			| Node {id,attrs} => prattrs clusters id attrs cont
			| GraphAttr attrs => prattrs clusters "graph" attrs cont
			| NodeAttr  attrs => prattrs clusters "node"  attrs cont
			| EdgeAttr  attrs => prattrs clusters "edge"  attrs cont
			| Subgraph gr     => 
				prgraph clusters gr ("subgraph " ::cont)

	and prstmts clusters stmts (cont :string list) : string list =
		foldl (prstmt clusters) cont stmts

	and prgraph clusters {cluster, id, stmts} (cont:string list) :string list = 
		let val pref  = if cluster then "\"cluster" else "\""
		    val cont1 = prstmts clusters stmts 
				("\" {\n" :: id :: pref :: cont) 
		in "}\n" :: cont1 end

	fun findclusters g :string list =
		let fun clg {cluster,id,stmts} = 
				if cluster then id :: cls stmts
				else cls stmts
		    and cls (Subgraph g::ss) = clg g @ cls ss
		      | cls (_::ss) = cls ss
		      | cls [] = []
		in clg g end

	fun prtopgraph ((strict, digraph, g),cont :string list) :string list = 
		let val clusters = findclusters g
		    val pref1 = if strict then "strict " else ""
		    val pref2 = if digraph then "digraph " else "graph "
		in prgraph clusters g (pref2 :: pref1 :: cont) end


	fun print (d :dot) :string list = 
		rev (foldl prtopgraph [] d)
				

end
